'use strict';

const crypto = require('crypto');
const fs = require('fs-extra');
const path = require('path');
const util = require('util');
const zlib = require('zlib');

const async = require('async');
const dir = require('node-dir');
const uglifyjs = require('uglify-js');
const CleanCss = require('clean-css');
const less = require('less');
const request = require('request');

const S3 = require('./s3');

const ONE_YEAR_IN_SECONDS = 31536000;
const THIRTY_DAYS_IN_SECONDS = 2592000;
const SEVEN_DAYS_IN_SECONDS = 604800;

/**
 * Constructs an AssetProcessor
 * @param {object} config
 * @constructor
 */
function AssetProcessor(config) {
    let baseRoot;
    let currRelativeRoot;

    this.config = config;

    if (config.s3) {
        this.s3 = new S3(config.s3.region, config.s3.bucket, config.s3.key, config.s3.secret);
        if (config.s3.cloudfrontMapping) {
            this.s3.setCloudfrontMapping(config.s3.cloudfrontMapping);
        }
    }
    
    // set default extensions for js, css, and images
    if (this.config.targets.javascripts && !this.config.targets.javascripts.extensions) {
        this.config.targets.javascripts.extensions = ['.js'];
    }
    if (this.config.targets.stylesheets && !this.config.targets.stylesheets.extensions) {
        this.config.targets.stylesheets.extensions = ['.css'];
    }
    if (this.config.targets.images && !this.config.targets.images.extensions) {
        this.config.targets.images.extensions = ['.ico', '.png','.jpg','.jpeg','.gif','.bmp','.svg'];
    }

    // figure out and normalize the relevant roots we are using

    if (!this.config.root) {
        throw new Error('Configuration root is required');
    } else if (!fs.existsSync(this.config.root)) {
        throw new Error('Invalid configuration root');
    }

    this.config.root = path.normalize(this.config.root); // full path that is the base root of the project; target roots are relative to this

    baseRoot = this.config.root;

    currRelativeRoot = this.config.targets.javascripts?.root; // javascript target root (if any)
    this.config.javascriptsRoot = path.normalize(currRelativeRoot ? path.resolve(baseRoot, currRelativeRoot) : baseRoot); // full path to the local root for javascipt assets

    currRelativeRoot = this.config.targets.stylesheets?.root; // stylesheet target root (if any)
    this.config.stylesheetsRoot = path.normalize(currRelativeRoot ? path.resolve(baseRoot, currRelativeRoot) : baseRoot); // full path to the local root for stylesheet assets

    currRelativeRoot = this.config.targets.images?.root; // image target root (if any)
    this.config.imagesRoot = path.normalize(currRelativeRoot ? path.resolve(baseRoot, currRelativeRoot) : baseRoot); // full path to the local root for image assets
    this.config.imagesRootRelative = path.relative(this.config.root, this.config.imagesRoot);

    currRelativeRoot = this.config.targets.extras?.root; // extra target root (if any)
    this.config.extrasRoot = path.normalize(currRelativeRoot ? path.resolve(baseRoot, currRelativeRoot) : baseRoot); // full path to the local root for extra assets
    this.config.extrasRootRelative = path.relative(this.config.root, this.config.extrasRoot);

    // Some other options
    this.forceCdnUpdate = this.config.forceCdnUpdate;
}

/**
 * Gets array of relevant javascript file paths based on the AssetProcessor's configuration
 * @param {function} callback function(err, files)
 */
AssetProcessor.prototype.getJavaScriptFiles = function (normalizedFullPath, callback) {
    getFilesRelativeToRoot(this.config.root, this.config.javascriptsRoot, this.config.targets.javascripts, normalizedFullPath, callback);
};

/**
 * Gets array of relevant javascript file paths based on the AssetProcessor's configuration
 * @param {function} callback function(err, files)
 */
AssetProcessor.prototype.getCssFiles = function (normalizedFullPath, callback) {
    getFilesRelativeToRoot(this.config.root, this.config.stylesheetsRoot, this.config.targets.stylesheets, normalizedFullPath, callback);
};

/**
 * Gets array of relevant image file paths based on the AssetProcessor's configuration
 * @param {function} callback function(err, files)
 */
AssetProcessor.prototype.getImageFiles = function (normalizedFullPath, callback) {
    getFilesRelativeToRoot(this.config.root, this.config.imagesRoot, this.config.targets.images, normalizedFullPath, callback);
};

/**
 * Gets array of relevant extra file paths based on the AssetProcessor's configuration
 * @param {function} callback function(err, files)
 */
AssetProcessor.prototype.getExtraFiles = function (normalizedFullPath, callback) {
    getFilesRelativeToRoot(this.config.root, this.config.extrasRoot, this.config.targets.extras, normalizedFullPath, callback);
};

/**
 * Uglifies and compresses JavaScript
 * @param excludeSourceMap
 * @param callback
 */
AssetProcessor.prototype.processJavaScript = function(excludeSourceMap, callback) {
    if (typeof excludeSourceMap === 'function') {
        callback = excludeSourceMap;
        excludeSourceMap = false;
    }

    async.auto({
        getFiles: [
            (next) => {
                this.getJavaScriptFiles(true, next);
            }
        ],
        getPath: [
            'getFiles',
            (next, results) => {
                getJavaScriptPath(results.getFiles, next);
            }
        ],
        minify: [
            'getPath',
            (next, results) => {
                const jsFiles = results.getFiles;
                let uglifyResult = {};
                let err;
                const mapPath = results.getPath.replace('.js','.map');
                const publicMapPath = this.s3 ? this.s3.urlWithBucket(mapPath) : mapPath;
                const uglifyOptions = excludeSourceMap ? {} : {
                    sourceMap: {
                        filename: results.getPath,
                        url: publicMapPath
                    }
                };

                if (!jsFiles.length) {
                    next(null, {});
                } else {
                    try {
                        const objectToMinify = {};
                        jsFiles.forEach(file => {
                            objectToMinify[file] = fs.readFileSync(file, 'utf8');
                        })
                        uglifyResult = uglifyjs.minify(objectToMinify, uglifyOptions);
                        if (uglifyResult.error) throw uglifyResult.error;
                    } catch (uglifyError) {
                        err = uglifyError;
                    }

                    next(err, { js: uglifyResult && uglifyResult.code, map: uglifyResult.map });
                }
            }
        ]
    }, callback);
}

/**
 * Saves processed JavaScript locally
 * @param callback
 * @param [outputDir] {string} directory to write the file (defaults to javascriptsRoot)
 */
AssetProcessor.prototype.saveJavaScriptToFile = function(callback, outputDir) {
    async.auto({
        getSource: [
            (next) => {
                this.processJavaScript(false, next);
            }
        ],
        saveFile: [
            'getSource',
            (next, results) => {
                let relativeUrl = '';
                const source = results.getSource.minify.js;

                if (source) {
                    const basePath = outputDir ? outputDir : this.config.javascriptsRoot;
                    const fullLocalPath = `${basePath}/${results.getSource.getPath}`;
                    relativeUrl = `/${results.getSource.getPath}`;

                    fs.outputFileSync(fullLocalPath, source);
                }

                next(null, relativeUrl);
            }
        ],
    }, (err, results) => {
        const result = {
            jsUrl: results && results.saveFile
        };
        callback(err, result);
    });

};

AssetProcessor.prototype.processCss = function(callback) {
    async.auto({
        getFiles: [
            (next) => {
                this.getCssFiles(true, next);
            }
        ],
        getPath: [
            'getFiles',
            (next, results) => {
                getCssPath(results.getFiles, next);
            }
        ],
        cleanCss: [
            'getFiles',
            (next, results) => {
                const files = results.getFiles;
                let css = '';

                async.eachSeries(files, (file, eachNext) => {
                    let currCss = ''+fs.readFileSync(file);
                    currCss = new CleanCss({noAdvanced: true, noRebase: true}).minify(currCss);
                    currCss = rebaseUrls(this, file, currCss); // we do our own special rebasing
                    css += `${currCss}\n`;
                    eachNext();
                }, (err) => {
                    next(err, css);
                });
            }
        ]
    }, callback);
};

/**
 * Saves processed css locally
 * @param callback
 * @param [outputDir] {string} directory to write the file (defaults to stylesheetsRoot)
 */
AssetProcessor.prototype.saveCssToFile = function(callback, outputDir) {
    async.auto({
        getSource: [
            (next) => {
                this.processCss(next)
            }
        ],
        saveFile: [
            'getSource',
            (next, results) => {
                let relativeUrl = '';
                const source = results.getSource.cleanCss;

                if (source) {
                    const basePath = outputDir ? outputDir : this.config.stylesheetsRoot;
                    const fullLocalPath = `${basePath}/${results.getSource.getPath}`;
                    relativeUrl = `/${results.getSource.getPath}`;

                    fs.outputFileSync(fullLocalPath, source);
                }

                next(null, relativeUrl);
            }
        ],
    }, (err, results) => {
        callback(err, {
            cssUrl: results && results.saveFile
        });
    });

};

/**
 * Imports latest stylesheets
 * @param callback {function} function(err, numImported)
 */
AssetProcessor.prototype.importLatestStylesheets = function(callback) {
    const gitToken = this.config.git?.token;
    const imports = this.config.targets?.stylesheets?.imports || {};

    const destDir = path.resolve(this.config.stylesheetsRoot, imports.destination || '');
    const sources = imports.sources || [];
    const mappings = imports.mappings || {};
    let numImported = 0;

    if (!fs.existsSync(destDir)) {
        console.log(`Creating import directory ${destDir}`);
        fs.mkdirsSync(destDir);
    }

    console.log(`Importing ${sources.length} css files to ${destDir}`);
    if (Object.keys(mappings).length) {
        console.log('Will be applying the following remappings:', mappings);
    }

    async.eachSeries(sources, (source, eachNext) => {
        const destPath = path.resolve(destDir, path.basename(source));
        console.log(`Importing ${source} to ${destPath}`);
        importFromGit(source, destPath, gitToken, (err) => {
            if (!err) {
                applyImportMappings(destPath, mappings, (err) => {
                    if (!err) {
                        numImported += 1;
                    }
                    eachNext(err);
                });
            } else {
                console.error(err);
                eachNext(err);
            }
        });
    }, (err) => {
        console.log(`Imported ${numImported} of ${sources.length} stylesheets`);
        callback(err, numImported);
    });
};

/**
 * Uglifies, compresses, and uploads the JavaScript to S3
 * @param {boolean=} excludeSourceMap [optional] if true source map will not be generated and uploaded
 * @param {function} callback function(err, uploadedUrl)
 */
 AssetProcessor.prototype.uploadJavaScriptToCdn = function(excludeSourceMap, callback) {
    if (typeof excludeSourceMap === 'function') {
        callback = excludeSourceMap;
        excludeSourceMap = false;
    }

    async.auto({
        getSource: [
            (next) => {
                this.processJavaScript(excludeSourceMap, next);
            }
        ],
        compressJs: [
            'getSource',
            (next, results) => {
                zlib.gzip(results.getSource.minify.js, next);
            }
        ],
        uploadJs: [
            'compressJs',
            async (next, results) => {
                const jsFiles = results.getSource.getFiles;
                const gzip = results.compressJs;
                const targetPath = results.getSource.getPath;
                const headers = {
                    'Content-Type':      'application/x-javascript',
                    'Content-Encoding':  'gzip',
                    'Content-Length':    gzip.length,
                    'Cache-Control':     'public, max-age=' + ONE_YEAR_IN_SECONDS
                };
                this.s3.putBufferCallback(gzip, targetPath, headers, next);
            }
        ],
        uploadMap: [
            'getSource',
            async (next, results) => {
                const mapPath = results.getSource.getPath.replace('.js','.map');
                if (!excludeSourceMap) {
                    this.s3.putBufferCallback(results.getSource.minify.map, mapPath, next);
                } else {
                    next();
                }
            }
        ],
        uploadCompleted: [
            'uploadJs',
            (next, results) => {
                const targetPath = results.getSource.getPath;
                next();
            }
        ]
    }, (err, results) => {
        callback(err, results && results.uploadJs);
    });
};

/**
 * Cleans up CSS, combines into one file, compresses it, and uploads it to S3
 * @param {function} callback function(err, cssUrl)
 */
 AssetProcessor.prototype.uploadCssToCdn = function(callback) {
    async.auto({
        getSource: [
            (next) => {
                this.processCss(next)
            }
        ],
        compressCss: [
            'getSource',
            (next, results) => {
                zlib.gzip(results.getSource.cleanCss, next);
            }
        ],
        uploadCss: [
            'compressCss',
            async (next, results) => {
                const cssFiles = results.getSource.getFiles;
                const gzip = results.compressCss;
                const targetPath = results.getSource.getPath;
                const headers = {
                    // 'x-amz-acl':         'public-read',
                    'Content-Type':      'text/css',
                    'Content-Encoding':  'gzip',
                    'Content-Length':    gzip.length,
                    'Cache-Control':     'public, max-age=' + ONE_YEAR_IN_SECONDS
                };
                this.s3.putBufferCallback(gzip, targetPath, headers, next);
            }
        ]
    }, (err, results) => {
        callback(err, results && results.uploadCss);
    });
};

/**
 * Uploads images to CDN, maintain same directory structure relative to image root
 * @param {function} callback function(err, imageFolderUri)
 */
 AssetProcessor.prototype.uploadImagesToCdn = function(callback) {
    async.auto({
        getFiles: [
            (next) => {
                this.getImageFiles(true, next);
            }
        ],
        getFolder: [
            (next) => {
                getImageFolder(this.config.imagesRootRelative, next);
            }
        ],
        uploadImages: [
            'getFiles',
            'getFolder',
            async (next, results) => {
                util.callbackify(uploadRelativeToRoot)(results.getFiles, this.config.imagesRoot, results.getFolder, this.s3, next)
            }
        ]
    }, (err, results) => {
        callback(err, results && results.uploadImages);
    });
};

/**
 * Uploads extras to CDN, maintain same directory structure relative to extras root
 * @param {function} callback function(err, imageFolderUri)
 */
 AssetProcessor.prototype.uploadExtrasToCdn = function(callback) {
    async.auto({
        getFiles: [
            (next) => {
                this.getExtraFiles(true, next);
            }
        ],
        getFolder: [
            (next) => {
                getExtrasFolder(this.config.extrasRootRelative, next);
            }
        ],
        uploadExtras: [
            'getFiles',
            'getFolder',
            async (next, results) => {
                util.callbackify(uploadRelativeToRoot)(results.getFiles, this.config.extrasRoot, results.getFolder, this.s3, next);
            }
        ]
    }, (err, results) => {
        callback(err, results && results.uploadExtras);
    });
};

/**
 * Ensures latest assets are processed and uploaded
 * @param {object} [opts] Options passed from the asset processor command
 * @param {boolean} [opts.skipJs] If true processing JavaScript files will be skipped
 * @param {function} callback function(err, results)
 * @return {object} result
 * @return {object} result.ensureJs
 * @return {object} result.ensureCss
 * @return {object} result.ensureImages
 * @return {object} result.ensureExtras
 */
AssetProcessor.prototype.ensureAssets = function(opts, callback) {
    if (typeof opts === 'function') {
        callback = opts;
        opts = {};
    }

    // note: we use artificial dependencies to limit concurrent S3 usage

    async.auto({
        ensureJs: [
            (next) => {
                if (!opts.skipJs) {
                    checkAndUpdateJavaScript(this, next);
                } else {
                    next();
                }
            }
        ],
        ensureCss: [
            'ensureJs',
            (next) => {
                checkAndUpdateCss(this, next);
            }
        ],
        ensureImages: [
            'ensureCss',
            (next) => {
                checkAndUpdateImages(this, next);
            }
        ],
        ensureExtras: [
        'ensureImages',
            (next) => {
                checkAndUpdateExtras(this, next);
            }
        ]
    }, (err, results) => {
        const result = {
            ...(results?.ensureJs ?? {}),
            ...(results?.ensureCss ?? {}),
            ...(results?.ensureImages ?? {}),
            ...(results?.ensureExtras ?? {}),
        }
        callback(err, result);
    });
};

/**
 * Helper function to upload files preserving directory structure within a relative root.
 * Calls back with the relative folder path it uploaded directory structure into.
 * @param {string[]} files array of full file paths
 * @param {string} activeRoot full path to the relevant root
 * @param {string} targetFolder path to upload to
 * @param {S3} s3 client doing the uploading
 * @throws {Error}
 * @private
 */
 async function uploadRelativeToRoot(files, activeRoot, targetFolder, s3) {
    for (const file of files) {
        const target = `${targetFolder}/${stripRoot(file, activeRoot)}`;
        const headers = {
            // 'x-amz-acl': 'public-read',
            'Cache-Control': 'public, max-age=' + SEVEN_DAYS_IN_SECONDS,
        };
        await s3.putFile(file, target, headers);        
    }

    return s3.urlWithBucket(targetFolder);
}

function getFilesRelativeToRoot(mainRoot, activeRoot, targets, normalizedFullPath, callback) {
    const relative = path.relative(mainRoot, activeRoot);

    if (typeof normalizedFullPath === 'function') {
        callback = normalizedFullPath;
        normalizedFullPath = false;
    }

    getFiles(targets, activeRoot, normalizedFullPath, (err, files) => {
        if (normalizedFullPath) {
            // if normalized, no need to modify paths
            callback(err, files);
        } else {
            // if not normalized, make files relative to main root
            // then replace windows "\\" path separator with "/"
            const relativeFiles = (files || []).map((file) => path.join(relative, file).replace(/\\/gi,'/'));
            callback(err, relativeFiles);
        }
    });
}

/**
 * Helper function that strips a root form the beginning of each file.
 * This will also convert file to use forward slashes
 * @param {string} file path of file from which to strip the root
 * @param {string} root the root path to strip away if file path begins with it
 * @returns {string} file path with the root stripped from each one
 */
function stripRoot(file, root) {
    const posixRoot = root && root.replace(/\\/g, '/');
    file = file.replace(/\\/g, '/');
    const index = file.indexOf(posixRoot);
    if (index !== -1) {
        file = file.substr(posixRoot.length);
        if (file[0] === '/') {
            // if we have a root and the file starts with slash, strip that too so path is relative
            file = file.substr(1);
        }
    }
    return  file;
}

/**
 * Helper function that uses a config paired with defaults to fall back on to get relevant files
 * @param {object} config the configuration for the file type to retrieve
 * @param {string} activeRoot to be used to strip out (or normalize)
 * @param {bool} normalizedFullPath [optional] if true, paths will be full paths (not relative to root) and expanded out (default: false)
 * @param {function} callback function(err, files)
 */
function getFiles(config, activeRoot, normalizedFullPath, callback) {
    if (typeof normalizedFullPath === 'function') {
        callback = normalizedFullPath;
        normalizedFullPath = false;
    }

    async.auto({
        getRelevantFiles: [
            (next) => {
                let directories = config.directories;
                let specificFiles = config.files;

                if (!directories && !specificFiles) {
                    directories = activeRoot;
                }
                if (directories && !(directories instanceof Array)) {
                    directories = [directories];
                }

                // need to resolve directory paths relative to root
                directories = (directories || []).map((directory) => {
                    return path.resolve(activeRoot, directory);
                });

                // need to resolve files relative to root
                specificFiles = (specificFiles || []).map((specificFile) => {
                    return path.resolve(activeRoot, specificFile);
                });

                const extensions = config.extensions || [];
                const pref = config.preference || [];
                const exclusions = config.exclude || [];

                getRelevantFiles(activeRoot, directories, specificFiles, extensions, pref, exclusions, next);
            }
        ],
        normalizeFullPathIfNeeded: [
            'getRelevantFiles',
            (next, results) => {
                let files = results.getRelevantFiles;
                if (normalizedFullPath) {
                    files = files.map((file) => {
                        if (file[0] !== '/') {
                            return path.normalize(path.join(activeRoot, file));
                        }
                    });
                }
                next(null, files);
            }
        ]
    }, (err, results) => {
        callback(err, results && results.normalizeFullPathIfNeeded);
    });
}

/**
 * Helper function that recursively searches a directory for relevant files based on any extension, preference, and exclusions.
 * @param {string|} root full path that is the root of the directories
 * @param {string|string[]} directories full path or array of paths to the directory to search; or null if root will be used.
 * @param {string|string[]} specificFiles full path or array of paths to specific files to include.
 * @param {string[]} extensions array of acceptable file extensions or null/empty if all are allowed
 * @param {string[]} pref array of file order that should be enforced or null/empty if no order
 * @param {string[]} exclusions  array of directories and files to exclude
 * @param {function} callback function(err, files)
 */
function getRelevantFiles(root, directories, specificFiles, extensions, pref, exclusions, callback) {
    async.auto({
        files: [
            (next) => {
                // recurse into all directories to get complete list of files
                let files = specificFiles || [];
                async.each(directories, (directory, eachNext) => {
                    dir.files(directory, (err, dirFiles) => {
                        files = files.concat(dirFiles || []);
                        eachNext(err);
                    });
                }, (err) => {
                    next(err, files);
                });
            }
        ],
        process: [
            'files',
            (next, results) => {
                // strip root of file (which will also ensure file path is in posix format)
                let files = results.files.map((file) => {
                    return stripRoot(file, root);
                });

                // filter the files
                files = files.filter((file) => {
                    // check against any extension requirements
                    const extMatch = extensionMatch(file, extensions);

                    // check against any exclusions
                    const excluded = (exclusions || []).some((excludePath) => {
                        return file.toLowerCase().indexOf(excludePath.toLowerCase()) === 0;
                    });

                    // check if the file is included in "preference"
                    const included = (pref || []).some((pref) => {
                        return file.toLowerCase() === pref.toLowerCase();
                    });

                    let include = false;
                    if (included) {
                        include = true;
                    } else if (extMatch) {
                        include = !excluded;
                    }

                    // if extension matches and not excluded, keep the file!
                    return include; 
                });

                // specific file orders will be kept if no preference
                pref = (pref || []).concat(specificFiles.map((specificFile) => {
                    return stripRoot(specificFile, root);
                }));

                // sort based on preference
                files.sort((a,b) => {
                    const preferenceDiff = preference(pref, a) - preference(pref, b);
                    return preferenceDiff !== 0 ? preferenceDiff : a < b ? -1 : a > b ? 1 : 0;
                });

                // sanity check against preference and specific file configuration
                pref.forEach((preferencePath) => {
                    const fullPath = path.resolve(root, preferencePath);
                    // don't want to warn about extensions we are filtering out
                    if (!fs.existsSync(fullPath) && (!path.extname(fullPath) || extensionMatch(fullPath, extensions))) {
                        console.warn(`Warning: ${preferencePath} was in configuration but cannot be found`);
                    }
                });

                next(null, files);
            }
        ]
    }, (err, results) => {
        callback(err, results && results.process);
    });
}

/**
 * Helper function that ranks files based on preference
 * @param {string[]} preferences  type of files to get (js|css)
 * @param {string} fileName the relative name of the file being ranked
 * @returns {number} the rank of the file taking into account order preference
 */
function preference(preferences, fileName) {
    let i;
    let currPreference;

    fileName = fileName.toLowerCase(); // Lets do case-insensitive file names

    for (i = 0; i < preferences.length; i += 1) {
        currPreference = preferences[i].toLowerCase();
        if (path.extname(currPreference)) {
            // preference is a file, lets see if we have an exact match
            if (currPreference === fileName) {
                break;
            }
        } else {
            // preference is a directory, see if its contained in it
            // we know the directory is contained if we never have to traverse up
            if (path.relative(currPreference, fileName).indexOf('..') === -1) {
                break;
            }
        }
    }

    return i;
}

/**
 * Computes a hash of the given files
 * @param {string[]} files array of full paths of the files to hash
 * @param {function} callback function(err, hash)
 */
function hashFiles(files, callback) {
    const hash = crypto.createHash('md5');

    async.eachSeries(files, (file, eachNext) => {
        fs.readFile(file, (err, data) => {
            hash.update(data);
            eachNext(err);
        });
    }, (err) => {
        callback(err, !err && files.length && hash.digest('hex'));
    });
}

function getJavaScriptPath(files, callback) {
    hashAndGeneratePath(files, 'js', '.js', callback);
}

function getCssPath(files, callback) {
    hashAndGeneratePath(files, 'css', '.css', callback);
}

function getImageFolder(imagesRootRelative, callback) {
    return firstDirNameFromRootPath(imagesRootRelative, 'img', callback);
}

function getExtrasFolder(extrasRootRelative, callback) {
    return firstDirNameFromRootPath(extrasRootRelative, 'extra', callback);
}

function firstDirNameFromRootPath(rootPath, defaultDirName, callback) {
    const folder = `${path.sep}${(firstDirName(rootPath) || defaultDirName)}`;
    if (callback) {
        callback(null, folder);
    }
    return folder;
}

function firstDirName(dirPath) {
    dirPath = (dirPath || '').replace(/[\\\/]/g, path.sep);

    if (dirPath[0] === path.sep) {
        dirPath = dirPath.substr(1);
    }

    return dirPath.split(path.sep)[0] || '';
}

function hashAndGeneratePath(files, path, extension, callback) {
    async.auto({
        hashFiles: [
            (next) => {
                hashFiles(files, next);
            }
        ],
        path: [
            'hashFiles',
            (next, results) => {
                const hash = results.hashFiles;
                const uri = `${path}/${hash}${extension}`;
                next(null, uri);
            }
        ]
    }, (err, results) => {
        callback(err, results && results.path);
    });
}

function checkAndUpdateJavaScript(self, callback) {
    async.auto({
        getJsFiles: [
            (next) => {
                self.getJavaScriptFiles(true, next);
            }
        ],
        getJsPath: [
            'getJsFiles',
            (next, results) => {
                getJavaScriptPath(results.getJsFiles, next);
            }
        ],
        checkJs: [
            'getJsPath',
            (next, results) => {
                self.s3.fileExistsCallback(results.getJsPath, next);
            }
        ],
        updateJsIfNeeded: [
            'checkJs',
            (next, results) => {
                const jsChanged = !results.checkJs;

                if (jsChanged || self.forceCdnUpdate) {
                    // time to update javascript
                    self.uploadJavaScriptToCdn((err, newUrl) => {
                        next(err, newUrl);
                    });
                } else {
                    // no update needed
                    next(null, self.s3.urlWithBucket(results.getJsPath));
                }
            }
        ]  
    }, (err, results) => {
        const result = {
            jsUrl: results && results.updateJsIfNeeded,
            jsChanged: results && !results.checkJs
        };
        callback(err, result);
    });
}

function checkAndUpdateCss(self, callback) {
    async.auto({
        getCssFiles: [
            (next) => {
                self.getCssFiles(true, next);
            }
        ],
        getCssPath: [
            'getCssFiles',
            (next, results) => {
                getCssPath(results.getCssFiles, next);
            }
        ],
        checkCss: [
            'getCssPath',
            (next, results) => {
                self.s3.fileExistsCallback(results.getCssPath, next);
            }
        ],
        updateCssIfNeeded: [
            'checkCss',
            (next, results) => {
                const cssChanged = !results.checkCss;

                if (cssChanged || self.forceCdnUpdate) {
                    // time to update css
                    self.uploadCssToCdn((err, newUrl) => {
                        next(err, newUrl);
                    });
                } else {
                    // no update needed
                    next(null, self.s3.urlWithBucket(results.getCssPath));
                }
            }
        ]
    }, (err, results) => {
        const result = {
            cssUrl: results && results.updateCssIfNeeded,
            cssChanged: results && !results.checkCss
        };
        callback(err, result);
    });
}

function checkAndUpdateImages(self, callback) {
    checkRelativeToRoot(self, 'image', callback);
}

function checkAndUpdateExtras(self, callback) {
    checkRelativeToRoot(self, 'extra', callback);
}

function checkRelativeToRoot(self, type, callback) {
    async.auto({
        getFiles: [
            (next) => {
                if (type === 'image') {
                    self.getImageFiles(true, next);
                } else {
                    self.getExtraFiles(true, next);
                }
            }
        ],
        getFolder: [
            (next) => {
                if (type === 'image') {
                    getImageFolder(self.config.imagesRootRelative, next);
                } else {
                    getExtrasFolder(self.config.extrasRootRelative, next);
                }
            }
        ],
        hashAndGeneratePath: [
            'getFiles',
            'getFolder', (next, results) => {
                hashAndGeneratePath(results.getFiles, results.getFolder, '.txt', next);
            }
        ],
        checkFiles: [
            'hashAndGeneratePath',
            (next, results) => {
                self.s3.fileExistsCallback(results.hashAndGeneratePath, next);
            }
        ],
        updateIfNeeded: [
            'checkFiles',
            (next, results) => {
                const filesChanged = !results.checkFiles;

                if (filesChanged || self.forceCdnUpdate) {
                    if (type === 'image') {
                        self.uploadImagesToCdn(next);
                    } else {
                        self.uploadExtrasToCdn(next);
                    }
                } else {
                    next(null, self.s3.urlWithBucket(results.getFolder));
                }
            }
        ],
        uploadIndicatorFileIfNeeded: [
            'updateIfNeeded',
            async (next, results) => {
                const targetPath = results.hashAndGeneratePath;
                if (!results.checkFiles) {
                    self.s3.putBufferCallback(results.getFiles.join('\n'), targetPath, next);
                } else {
                    next();
                }
            }
        ]
    }, (err, results) => {
        const result = {};

        // we use the plural of type for results
        type = `${type}s`;

        result[`${type}Url`] = results && results.updateIfNeeded;
        result[`${type}Changed`] = !results.checkFiles;

        callback(err, result);
    });
}

/**
 * Rebases urls found within url() in our CSS to be relative to the stylesheet. This
 * allows our assets to be properly linked no matter where they are hosted.
 * 
 * @param {object} self the "this" scope of an AssetProcessor
 * @param {string} file the full path to the css file (used to help resolve relative paths)
 * @param {string} css the css to process
 * @returns {string} the css with url() functions rebased
 */
function rebaseUrls(self, file, css) {
    const cssRoot = self.config.stylesheetsRoot;
    const extrasRoot = self.config.extrasRoot;
    const imagesRoot = self.config.imagesRoot;

    let normalizedUrl; // url normalized to a file system path
    let assetPath; // full path of the asset
    let rebasedUrl; // rebased url

    // Regular expressions used to test each url found
    const urlRegex = /url\((['"]?([^'"\)]+)['"]?)\)/ig;
    const externalHttpRegex = /^(http(s)?:)?\/\//i;
    const absoluteUrlRegex = /^[\/\\][^\/]/;
    const base64Regex = /^data:.+;base64/i;

    let match;
    let url;
    let startIndex;
    let endIndex;

    file = path.normalize(file);

    while ((match = urlRegex.exec(css)) !== null) {
        url = match[1];
        startIndex = match.index+match[0].indexOf(url); // we want start index of just the actual url
        endIndex = startIndex + url.length;

        // clean off any quotes from url, as they are not needed
        url = url.replace(/^['"]/, '').replace(/['"]$/, '');

        // we don't want to rebase any external urls, so first check for that
        if (!externalHttpRegex.test(url) && !base64Regex.test(url)) {

            // if here, we are referencing our own files. Lets see if its an image or something extra so we know which root to use.
            normalizedUrl = path.normalize(url);

            // the asset path will either be absolute location from the root, or relative to the file
            if (absoluteUrlRegex.test(normalizedUrl)) {
                /**
                 * If the path is abosolute we can leave it as is, since it is already
                 * a web safe url. 
                 * 
                 * -- Example --
                 * Asset in CSS: '/images/logo.png'
                 */
                rebasedUrl = normalizedUrl;

            } else {
                /**
                 * If we have a relative asset in our CSS we are going to resolve it 
                 * to its local path and then make it an absolute web path based on 
                 * the css root path.
                 * 
                 * -- Example --
                 * CSS: /app/public/stylesheet/styles.css
                 * Asset Path in CSS: ../images/logo.png
                 * Resolved Local Path: /app/public/images/logo.png
                 * Resolved Web Path: /images/logo.png
                 */
                rebasedUrl = path
                    .resolve(path.dirname(file), normalizedUrl)
                    .replace(cssRoot, '')
                    .replace(extrasRoot, path.basename(extrasRoot))
                    .replace(imagesRoot, path.basename(imagesRoot));

                // Ensure rebase is not relative
                if (rebasedUrl[0] != '/') {
                    rebasedUrl = `/${rebasedUrl}`;
                }

            }

            // replace url with rebased one we do so by splicing out match and splicing in url; replacement functions could inadvertently replace repeat urls
            css = css.substr(0, startIndex) + rebasedUrl + css.substr(endIndex);

            // since we've modified the string, indexes may have change. have regex resume searching at end of replaced url
            urlRegex.lastIndex = endIndex;
        }
    }

    return css;
}

/**
 * Imports a file from a github repo to the file system
 * @param sourceUrl {string} url to the file to import
 * @param destPath {string} path to save the file
 * @param accessToken {string} github access token used to request file
 * @param callback {function} function(err, importSuccess)
 */
function importFromGit(sourceUrl, destPath, accessToken, callback) {
    const apiUrl = toGitHubApiSource(sourceUrl);
    const requestOptions = {
        url: apiUrl,
        headers: {
            Authorization: `token ${accessToken}`,              // our authentication
            Accept: 'application/vnd.github.v3.raw',          // we want raw response
            'User-Agent': 'NodeBot (+http://videoblocks.com)' // need this or github api will reject us
        }
    };

    const ws = fs.createWriteStream(destPath);

    // Create the streams
    const req = request(requestOptions);
    let statusCode;
    let reqErr;

    req.on('response', (response) => {
        statusCode = response.statusCode;
    });

    req.on('end', () => {
        const success = statusCode === 200;
        reqErr = reqErr || (!success ? new Error(`Error requesting ${apiUrl}; status code ${statusCode}`) : null);
    });

    req.on('error', (err) => {
        reqErr = reqErr || err;
    });

    ws.on('close', () => {
        const success = statusCode === 200;
        callback(reqErr, success);
    });

    req.pipe(ws);
}

function applyImportMappings(filePath, mappings, callback) {
    async.auto({
        readFile: [
            (next) => {
                fs.readFile(filePath, 'utf8', next);
            }
        ],
        replaceMappings: [
            'readFile',
                (next, results) => {
                let css = results.readFile;

                Object.keys(mappings || {}).forEach((importedRoot) => {
                    const search = escapeRegExp(importedRoot);
                    const replacement = mappings[importedRoot];
                    css = css.replace(new RegExp(search, 'g'), replacement);
                });

                next(null, css);
            }
        ],
        writeFile: [
            'replaceMappings',
            (next, results) => {
                const remappedCss = results.replaceMappings;
                fs.writeFile(filePath, remappedCss, next);
            }
        ]
    }, (err) => {
        callback(err);
    });
}

/**
 * Changes web urls for files on github to be compliant with the github API
 * Examples:
 *     https://github.com/Footage-Firm/StockBlocks/blob/master/assets/common/stylesheets/admin/on_off.css ->
 *         https://api.github.com/repos/Footage-Firm/StockBlocks/contents/assets/common/stylesheets/admin/on_off.css
 *
 * @param sourceUrl {string}
 */
function toGitHubApiSource(sourceUrl) {
    // web url is of format http://github.com/<org>/<repo>/blob/<branch>/<path>
    const gitHubBlobUrlRegex = /^https?:\/\/github\.com\/([^\/]+)\/([^\/]+)\/blob\/[^\/]+\/(.*)$/;
    const match = gitHubBlobUrlRegex.exec(sourceUrl);
    let org;
    let repo;
    let path;
    if (match) {
        // change to api format
        org = match[1];
        repo = match[2];
        path = match[3];
        sourceUrl = `https://api.github.com/repos/${org}/${repo}/contents/${path}`;
    }

    if (!sourceUrl.match(/^https?:\/\/api\.github\.com\/repos\//)) {
        throw new Error(`Invalid github import source URL: ${sourceUrl}`);
    }

    return sourceUrl;
}

/**
 * Helper function to see if a path points to a file whose extension (case-insensitive) matches one of many possibilities.
 * If no extensions provided its considered a wildcard match.
 * @param {string} filePath path to the file
 * @param {string[]} extensions the possible extensions
 * @returns {boolean} whether or not an extension matched
 */
function extensionMatch(filePath, extensions) {
    filePath = (filePath || '').toLowerCase();
    return !extensions || !extensions.length || extensions.some(e => path.extname(filePath).toLowerCase() === e.toLowerCase());
}

function escapeRegExp(string) {
    return string.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1");
}

module.exports = AssetProcessor;
