'use strict';

const fs = require('fs');
const path = require('path');
const { S3Client, HeadObjectCommand, PutObjectCommand } = require("@aws-sdk/client-s3");
const mime = require('mime');
const util = require('util');

// Set up our custom mime times for fonts, the second argument tells mime to override existing definitions
mime.define({
    'application/x-font-opentype': ['otf','eot','ttf'],
    'image/svg+xml': ['svg'],
    'application/font-woff': ['woff']
}, true);

/**
 * Constructs a wrapped S3 client for the given bucket and credentials
 * @param {string} region
 * @param {string} bucket
 * @param {string} key
 * @param {string} secret
 * @constructor
 */
function S3(region, bucket, key, secret) {
    const s3Config = {
        region,
        httpOptions: { timeout: 180000 },
        maxAttempts: 3,
    };
    if (key) {
        s3Config.credentials = {
            accessKeyId: key,
            secretAccessKey: secret,
        }
    }

    this.s3Client = new S3Client(s3Config);
    this.bucket = bucket;
}

/**
 * Enables mapping of bucket to cloudfront.
 * If value is non-null, urlWithBucket() will now map to a cloudfront url
 * @param cloudfrontMapping the cloudfront subdomain the S3 bucket is mapped to
 */
S3.prototype.setCloudfrontMapping = function(cloudfrontMapping){
    this.cloudfrontMapping = cloudfrontMapping;
};

/**
 * Gets the bucket in use by the S3 client
 * @returns {string} the name of the bucket in use
 */
S3.prototype.getBucket = function() {
    return this.bucket;
};

/**
 * Uploads the given text contents to S3
 * @param {string|Buffer} contents the text or buffer to upload
 * @param {string} targetPath the path the contents should be saved to
 * @param {bool|object=} makePrivateOrHeaders whether or not to make the file private on S3; or the full set of headers to use.
 * @returns {Promise<string>}
 * @throws {Error}
 */
S3.prototype.putBuffer = async function(contents, targetPath, makePrivateOrHeaders) {
    targetPath = cleanPath(targetPath);

    const buffer = contents instanceof Buffer ? contents : Buffer.from(contents);
    const result = await uploadDataToS3(this.s3Client, buffer, buffer.length, this.bucket, targetPath, makePrivateOrHeaders);

    return result && this.urlWithBucket(targetPath);
};

/**
 * Same as putBuffer, but takes an error-first style callback instead of returning a promise
 * @param {string|Buffer} contents the text or buffer to upload
 * @param {string} targetPath the path the contents should be saved to
 * @param {bool|object} makePrivateOrHeaders whether or not to make the file private on S3; or the full set of headers to use.
 * @param {function} callback 
 * @returns {void}
 */
S3.prototype.putBufferCallback = function(contents, targetPath, makePrivateOrHeaders, callback) {
    if (typeof makePrivateOrHeaders === 'function') {
        callback = makePrivateOrHeaders;
        makePrivateOrHeaders = null;
    }

    const callbackFn = util.callbackify(this.putBuffer.bind(this));
    callbackFn(contents, targetPath, makePrivateOrHeaders, callback);
}

/**
 * Uploads a file to S3
 * @param {string} sourcePath the path of the file to upload
 * @param {string} targetPath the path of where the file should be saved on S3
 * @param {bool|object} makePrivateOrHeaders whether or not to make the file private on S3; or the full set of heders to use.
 * @returns {Promise<string>}
 * @throws {Error}
 */
S3.prototype.putFile = async function(sourcePath, targetPath, makePrivateOrHeaders) {
    targetPath = cleanPath(targetPath);

    const stream = fs.createReadStream(sourcePath);
    const result = await uploadDataToS3(this.s3Client, stream, fs.statSync(sourcePath).size, this.bucket, targetPath, makePrivateOrHeaders);

    return result && this.urlWithBucket(targetPath);
};

/**
 * Same as putFile, but takes an error-first style callback instead of returning a promise
 * @param {string} sourcePath the path of the file to upload
 * @param {string} targetPath the path of where the file should be saved on S3
 * @param {bool|object} makePrivateOrHeaders whether or not to make the file private on S3; or the full set of heders to use.
 * @param {function} callback
 * @returns {void}
 */
S3.prototype.putFileCallback = function(sourcePath, targetPath, makePrivateOrHeaders, callback) {
    if (typeof makePrivateOrHeaders === 'function') {
        callback = makePrivateOrHeaders;
        makePrivateOrHeaders = null;
    }

    const callbackFn = util.callbackify(this.putFile.bind(this));
    callbackFn(sourcePath, targetPath, makePrivateOrHeaders, callback);
}

/**
 * Checks whether or not a file exists on S3
 * @param {string} path the path within the S3 bucket to check
 * @returns {Promise<boolean>}
 */
S3.prototype.fileExists = async function(path) {
    path = cleanPath(path);

    const params = {
        Bucket: this.bucket,
        Key: path,
    };

    try {
        const result = await this.s3Client.send(new HeadObjectCommand(params));
        return parseInt(result.ContentLength, 10) >= 0;
    } catch (e) {
        // only throw an error if it's something other than not being found
        if (e.name !== 'NotFound') {
            throw new Error(e);
        }
    }
};

/**
 * Same as fileExists, but takes an error-first style callback instead of returning a promise
 * @param {string} path the path within the S3 bucket to check
 * @param {function} callback 
 * @returns {void}
 */
S3.prototype.fileExistsCallback = function(path, callback) {
    const callbackFn = util.callbackify(this.fileExists.bind(this));
    callbackFn(path, callback);
}

/**
 * Returns the full url given an S3 path
 * @param {string} path path within the S3 bucket
 * @returns {string} url to the s3 path (or mapped cloudfront path)
 */
S3.prototype.urlWithBucket = function(path) {
    path = path || '';
    // for the web, forward slashes are king
    path = path.replace(/\\/g, '/');
    if (path[0] === '/') {
        path = path.substr(1);
    }
    return this.cloudfrontMapping
        ? 'https://'+this.cloudfrontMapping+'.cloudfront.net/'+path
        : 'https://s3.amazonaws.com/'+this.bucket.toLowerCase()+'/'+path;
};

/**
 * Uploads data to S3
 * @param {S3Client} s3Client 
 * @param {Buffer|fs.ReadStream} data 
 * @param {number} contentLength 
 * @param {string} bucket 
 * @param {string} targetPath 
 * @param {bool|object} makePrivateOrHeaders 
 * @returns {Promise<AWS.S3.ManagedUpload.SendData>}
 * @throws {Error}
 */
async function uploadDataToS3(s3Client, data, contentLength, bucket, targetPath, makePrivateOrHeaders) {
    let params = {
        Bucket: bucket,
        Key: targetPath,
        Body: data,
        ContentLength: contentLength,
    };

    params = extendParamsWithHeaders(params, makePrivateOrHeaders, targetPath);

    return s3Client.send(new PutObjectCommand(params));
}

function extendParamsWithHeaders(params, makePrivateOrHeaders, targetPath) {
    let headers = {'x-amz-acl': 'private'}

    if (typeof makePrivateOrHeaders === 'object') {
        headers = { ...headers, ...makePrivateOrHeaders };
    } else if (typeof makePrivateOrHeaders === 'boolean') {
        headers['x-amz-acl'] = makePrivateOrHeaders ? 'private' : 'public-read';
    }

    // some intelligent header setting based on the path
    if (!headers['Content-Type'] && targetPath) {
        headers['Content-Type'] = mime.getType(targetPath);
    }

    for (const [key, value] of Object.entries(headers)) {
        if (key === 'x-amz-acl') {
            params.ACL = value;
        } else if (key === 'Content-Type') {
            params.ContentType = value;
        } else if (key === 'Content-Encoding') {
            params.ContentEncoding = value;
        } else if (key === 'Content-Length') {
            params.ContentLength = value;
        } else if (key === 'x-amz-storage-class') {
            params.StorageClass = value;
        } else if (key === 'Cache-Control') {
            params.CacheControl = value;
        } else if (key === 'Expires') {
            params.Expires = value;
        } else {
            throw new Error(`Unsupported header ${key} for s3 sdk option mapping`);
        }
    }

    return params;
}

function cleanPath(path) {
    // cleans path in case generated in windows environment, and strips leading backslash if necessary
    return path && path.replace(/\\/g,'/').replace(/^\//,'');
}

module.exports = S3;